@@ -106,7 +106,7 @@ getpmparameter(UNUSED(HashTable ht), const char *name)
 	pm->u.str = paramtypestr(rpm);
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -224,7 +224,7 @@ getpmcommand(UNUSED(HashTable ht), const char *name)
 	}
     } else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -289,7 +289,7 @@ setfunction(char *name, char *val, int dis)
     shf = (Shfunc) zshcalloc(sizeof(*shf));
     shf->funcdef = dupeprog(prog, 0);
     shf->node.flags = dis;
-    shf->emulation = sticky_emulation;
+    shfunc_set_sticky(shf);
 
     if (!strncmp(name, "TRAP", 4) &&
 	(sn = getsignum(name + 4)) != -1) {
@@ -392,25 +392,41 @@ getfunction(UNUSED(HashTable ht), const char *name, int dis)
 				((shf->node.flags & PM_TAGGED) ? "t" : "")));
 	} else {
 	    char *t = getpermtext(shf->funcdef, NULL, 1), *n, *h;
+	    char *start;
+
+	    if (shf->redir)
+		start = "{\n\t";
+	    else
+		start = "\t";
 
 	    if (shf->funcdef->flags & EF_RUN) {
 		n = nicedupstring(name);
-		h = (char *) zhalloc(strlen(t) + strlen(n) + 9);
-		h[0] = '\t';
-		strcpy(h + 1, t);
+		h = (char *) zhalloc(strlen(start) + strlen(t) + strlen(n) + 8);
+		strcpy(h, start);
+		strcat(h, t);
 		strcat(h, "\n\t");
 		strcat(h, n);
 		strcat(h, " \"$@\"");
 	    } else
-		h = dyncat("\t", t);
+		h = dyncat(start, t);
 	    zsfree(t);
+	    /*
+	     * TBD: Is this unmetafy correct?  Surely as this
+	     * is a parameter value it stays metafied?
+	     */
 	    unmetafy(h, NULL);
 
+	    if (shf->redir) {
+		t = getpermtext(shf->redir, NULL, 1);
+		h = zhtricat(h, "\n}", t);
+		zsfree(t);
+	    }
+
 	    pm->u.str = h;
 	}
     } else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -456,21 +472,38 @@ scanfunctions(UNUSED(HashTable ht), ScanFunc func, int flags, int dis)
 				    ((shf->node.flags & PM_TAGGED) ? "Ut" : "U") :
 				    ((shf->node.flags & PM_TAGGED) ? "t" : "")));
 		    } else {
-			char *t = getpermtext(((Shfunc) hn)->funcdef, NULL, 1);
-			char *n;
+			Shfunc shf = (Shfunc)hn;
+			char *t = getpermtext(shf->funcdef, NULL, 1);
+			char *n, *start;
 
-			if (((Shfunc) hn)->funcdef->flags & EF_RUN) {
+			if (shf->redir)
+			    start = "{\n\t";
+			else
+			    start = "\t";
+
+			if (shf->funcdef->flags & EF_RUN) {
 			    n = nicedupstring(hn->nam);
-			    pm.u.str = (char *) zhalloc(strlen(t) + strlen(n) + 9);
-			    pm.u.str[0] = '\t';
-			    strcpy(pm.u.str + 1, t);
+			    pm.u.str = (char *) zhalloc(
+				strlen(start) + strlen(t) + strlen(n) + 8);
+			    strcpy(pm.u.str, start);
+			    strcat(pm.u.str, t);
 			    strcat(pm.u.str, "\n\t");
 			    strcat(pm.u.str, n);
 			    strcat(pm.u.str, " \"$@\"");
 			} else
-			    pm.u.str = dyncat("\t", t);
+			    pm.u.str = dyncat(start, t);
+			/*
+			 * TBD: Is this unmetafy correct?  Surely as this
+			 * is a parameter value it stays metafied?
+			 */
 			unmetafy(pm.u.str, NULL);
 			zsfree(t);
+
+			if (shf->redir) {
+			    t = getpermtext(shf->redir, NULL, 1);
+			    pm.u.str = zhtricat(pm.u.str, "\n}", t);
+			    zsfree(t);
+			}
 		    }
 		}
 		func(&pm.node, flags);
@@ -661,7 +694,7 @@ getbuiltin(UNUSED(HashTable ht), const char *name, int dis)
 	pm->u.str = dupstring(t);
     } else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -759,6 +792,38 @@ disreswordsgetfn(UNUSED(Param pm))
     return getreswords(DISABLED);
 }
 
+/* Functions for the patchars special parameter. */
+
+/**/
+static char **
+getpatchars(int dis)
+{
+    int i;
+    char **ret, **p;
+
+    p = ret = (char **) zhalloc(ZPC_COUNT * sizeof(char *));
+
+    for (i = 0; i < ZPC_COUNT; i++)
+	if (zpc_strings[i] && !dis == !zpc_disables[i])
+	    *p++ = dupstring(zpc_strings[i]);
+
+    *p = NULL;
+
+    return ret;
+}
+
+static char **
+patcharsgetfn(UNUSED(Param pm))
+{
+    return getpatchars(0);
+}
+
+static char **
+dispatcharsgetfn(UNUSED(Param pm))
+{
+    return getpatchars(1);
+}
+
 /* Functions for the options special parameter. */
 
 /**/
@@ -771,7 +836,7 @@ setpmoption(Param pm, char *value)
 	zwarn("invalid value: %s", value);
     else if (!(n = optlookup(pm->node.nam)))
 	zwarn("no such option: %s", pm->node.nam);
-    else if (dosetopt(n, (value && strcmp(value, "off")), 0))
+    else if (dosetopt(n, (value && strcmp(value, "off")), 0, opts))
 	zwarn("can't change option: %s", pm->node.nam);
     zsfree(value);
 }
@@ -784,7 +849,7 @@ unsetpmoption(Param pm, UNUSED(int exp))
 
     if (!(n = optlookup(pm->node.nam)))
 	zwarn("no such option: %s", pm->node.nam);
-    else if (dosetopt(n, 0, 0))
+    else if (dosetopt(n, 0, 0, opts))
 	zwarn("can't change option: %s", pm->node.nam);
 }
 
@@ -812,7 +877,7 @@ setpmoptions(UNUSED(Param pm), HashTable ht)
 	    if (!val || (strcmp(val, "on") && strcmp(val, "off")))
 		zwarn("invalid value: %s", val);
 	    else if (dosetopt(optlookup(hn->nam),
-			      (val && strcmp(val, "off")), 0))
+			      (val && strcmp(val, "off")), 0, opts))
 		zwarn("can't change option: %s", hn->nam);
 	}
     deleteparamtable(ht);
@@ -844,7 +909,7 @@ getpmoption(UNUSED(HashTable ht), const char *name)
     }
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -902,7 +967,7 @@ getpmmodule(UNUSED(HashTable ht), const char *name)
 	pm->u.str = dupstring(type);
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -1016,7 +1081,7 @@ getpmhistory(UNUSED(HashTable ht), const char *name)
 	pm->u.str = dupstring(he->node.nam);
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -1126,7 +1191,7 @@ getpmjobtext(UNUSED(HashTable ht), const char *name)
 	pm->u.str = pmjobtext(job);
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -1227,7 +1292,7 @@ getpmjobstate(UNUSED(HashTable ht), const char *name)
 	pm->u.str = pmjobstate(job);
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -1293,7 +1358,7 @@ getpmjobdir(UNUSED(HashTable ht), const char *name)
 	pm->u.str = pmjobdir(job);
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -1419,7 +1484,7 @@ getpmnameddir(UNUSED(HashTable ht), const char *name)
 	pm->u.str = dupstring(nd->dir);
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -1470,7 +1535,7 @@ getpmuserdir(UNUSED(HashTable ht), const char *name)
 	pm->u.str = dupstring(nd->dir);
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -1722,7 +1787,7 @@ getalias(HashTable alht, UNUSED(HashTable ht), const char *name, int flags)
 	pm->u.str = dupstring(al->text);
     else {
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     }
     return &pm->node;
 }
@@ -1918,7 +1983,7 @@ getpmusergroups(UNUSED(HashTable ht), const char *name)
     if (!gs) {
 	zerr("failed to retrieve groups for user: %e", errno);
 	pm->u.str = dupstring("");
-	pm->node.flags |= PM_UNSET;
+	pm->node.flags |= (PM_UNSET|PM_SPECIAL);
 	return &pm->node;
     }
 
@@ -1933,7 +1998,7 @@ getpmusergroups(UNUSED(HashTable ht), const char *name)
     }
 
     pm->u.str = dupstring("");
-    pm->node.flags |= PM_UNSET;
+    pm->node.flags |= (PM_UNSET|PM_SPECIAL);
     return &pm->node;
 }
 
@@ -2018,6 +2083,10 @@ static const struct gsu_array reswords_gsu =
 { reswordsgetfn, arrsetfn, stdunsetfn };
 static const struct gsu_array disreswords_gsu =
 { disreswordsgetfn, arrsetfn, stdunsetfn };
+static const struct gsu_array patchars_gsu =
+{ patcharsgetfn, arrsetfn, stdunsetfn };
+static const struct gsu_array dispatchars_gsu =
+{ dispatcharsgetfn, arrsetfn, stdunsetfn };
 static const struct gsu_array dirs_gsu =
 { dirsgetfn, dirssetfn, stdunsetfn };
 static const struct gsu_array historywords_gsu =
@@ -2038,6 +2107,8 @@ static struct paramdef partab[] = {
 	    &pmdisfunctions_gsu, getpmdisfunction, scanpmdisfunctions),
     SPECIALPMDEF("dis_galiases", 0,
 	    &pmdisgaliases_gsu, getpmdisgalias, scanpmdisgaliases),
+    SPECIALPMDEF("dis_patchars", PM_ARRAY|PM_READONLY,
+	    &dispatchars_gsu, NULL, NULL),
     SPECIALPMDEF("dis_reswords", PM_ARRAY|PM_READONLY,
 	    &disreswords_gsu, NULL, NULL),
     SPECIALPMDEF("dis_saliases", 0,
@@ -2072,6 +2143,8 @@ static struct paramdef partab[] = {
 	    &pmoptions_gsu, getpmoption, scanpmoptions),
     SPECIALPMDEF("parameters", PM_READONLY,
 	    NULL, getpmparameter, scanpmparameters),
+    SPECIALPMDEF("patchars", PM_ARRAY|PM_READONLY,
+	    &patchars_gsu, NULL, NULL),
     SPECIALPMDEF("reswords", PM_ARRAY|PM_READONLY,
 	    &reswords_gsu, NULL, NULL),
     SPECIALPMDEF("saliases", 0,
